use std::{cell::RefCell, sync::Arc};

use dbus::{Path, ffidisp::Connection};
use dbus_tree::{self as tree, Interface, MTFn};

mod mytest;

#[derive(Debug)]
struct Device {
    pub name: RefCell<String>,
    pub path: Path<'static>,
}

impl Device {
    pub fn new(index: u32, name: String) -> Self {
        Self {
            name: RefCell::new(name),
            path: format!("/Device{}", index).into(),
        }
    }
}

impl mytest::ComExampleMytest for Device {
    fn hello(&self, name: &str) -> Result<String, tree::MethodErr> {
        Ok(format!("Hello, {}!", name))
    }

    fn name(&self) -> Result<String, tree::MethodErr> {
        let name = self.name.borrow();

        Ok(name.clone())
    }

    fn setname(&self, value: String) -> Result<(), tree::MethodErr> {
        self.name.replace(value);
        Ok(())
    }
}

#[derive(Copy, Clone, Default, Debug)]
struct TData;
impl tree::DataType for TData {
    type Tree = ();
    type ObjectPath = Arc<Device>;
    type Property = ();
    type Interface = ();
    type Method = ();
    type Signal = ();
}

fn create_iface() -> Interface<MTFn<TData>, TData> {
    let f = tree::Factory::new_fn();
    mytest::com_example_mytest_server(&f, (), |m| {
        // Just provide a link from MethodInfo (m) to the &Device
        // we should call.
        let a: &Arc<Device> = m.path.get_data();
        let b: &Device = &a;
        b
    })
}

fn create_tree(
    devices: &[Arc<Device>],
    iface: &Arc<Interface<MTFn<TData>, TData>>,
) -> tree::Tree<MTFn<TData>, TData> {
    let f = tree::Factory::new_fn();
    let mut tree = f.tree(());
    for dev in devices {
        tree = tree.add(
            f.object_path(dev.path.clone(), dev.clone())
                .introspectable()
                .add(iface.clone()),
        );
    }
    tree
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let device = Arc::new(Device::new(0, format!("Device{}", 0)));

    // Create tree
    let iface = create_iface();
    let tree = create_tree(&[device], &Arc::new(iface));

    let c = Connection::new_session()?;
    c.register_name("com.example.mytest", 0)?;
    tree.set_registered(&c, true)?;

    c.add_handler(tree);
    loop {
        c.incoming(1000).next();
    }
}
